C       10        20        30        40        50        60        70
C23456789_C23456789_C23456789_C23456789_C23456789_C23456789_C23456789_

C *******************************************************************
C *                                                                 *
C *  Program:  struct_1d.f                                          *
C *                                                                 *
C *  Author:   Mariusz Zaczek  (zaczek@uiuc.edu)                    *
C *  Date:     May 7,2000                                           *
C *                                                                 *
C *******************************************************************

       program struct_1d
       implicit none

       integer max_nodes,max_elem
       parameter(max_nodes=101,max_elem=100)

       integer numsteps,step_incr,numelem,numbcs, 
     >         imposed_type(max_nodes),imposed_node(max_nodes)

       real*8 delta_t,length(max_elem),density(max_elem),
     >        E(max_elem),area(max_elem),beta,gamma,R(max_nodes),
     >        K(max_nodes,max_nodes),M(max_nodes,max_nodes),
     >        BplusY,BminusY,delta_sq_inv,imposed_Fd(max_nodes),
     >        Full_left(max_nodes,max_nodes),traction(max_elem),
     >        Right_cur(max_nodes,max_nodes),
     >        Right_pre(max_nodes,max_nodes)

C      ! call subroutine 'file_input' to read input data from file
       call  file_input( numsteps,step_incr,delta_t,beta,gamma, 
     >       length,density,E,area,numelem,numbcs,imposed_node,  
     >       imposed_type,imposed_Fd,traction )

C      ! Initialize some constants used in the calculations
       BplusY = 0.5 - 2.0*beta + gamma
       BminusY = 0.5 + beta - gamma
       delta_sq_inv = 1.0/(delta_t*delta_t)

C      ! assemble the [K],[M] matrices and {R} vector
       call  assemble( K,M,R,E,area,length,density,numelem,BplusY,
     >       BminusY,delta_sq_inv,Full_left,Right_cur,Right_pre,
     >       numbcs,imposed_node,imposed_type,imposed_Fd,
     >       traction,beta  )  

C      ! Begin solution loop
       call  solution( delta_t,beta,gamma,K,M,R,E,area,length,
     >       density,numelem,BplusY,BminusY,delta_sq_inv,
     >       Full_left,Right_cur,Right_pre,numsteps,
     >       numbcs,imposed_node,imposed_type,imposed_Fd,
     >       step_incr,traction )

       end    ! End of struct_1d program


C *******************************************************************
C   
C   solution
C 
C *******************************************************************
       subroutine solution( delta_t,beta,gamma,K,M,R,E,area,length,
     >            density,numelem,BplusY,BminusY,delta_sq_inv,
     >            Full_left,Right_cur,Right_pre,numsteps,
     >            numbcs,imposed_node,imposed_type,imposed_Fd,
     >            step_incr,traction )

       integer max_nodes,max_elem
       parameter(max_nodes=101,max_elem=100)

       integer i,j,numelem,numsteps,step_incr,x,numbcs,
     >         imposed_node(max_nodes),imposed_type(max_nodes),
     >         indx(max_nodes)
	integer y

       real*8 delta_t,D(max_nodes),BplusY,BminusY,beta,gamma,
     >        delta_sq_inv,R(max_nodes),K(max_nodes,max_nodes),
     >        D_old(max_nodes),M(max_nodes,max_nodes),b(max_nodes),
     >        Full_left(max_nodes,max_nodes),d_info,
     >        Right_cur(max_nodes,max_nodes),E(max_elem),
     >        Right_pre(max_nodes,max_nodes),area(max_elem),
     >        A(max_nodes,max_nodes),imposed_Fd(max_nodes),
     >        length(max_elem),density(max_elem),traction(max_elem)

C      ! Initialize displacements and {b} vector to zero 
       do i=1,numelem+1
             D(i) = 0.0
         D_old(i) = 0.0
	     b(i) = 0.0
       enddo

       write(99,*) "Matrix: Full_left"
       do y=1,numelem+1
         write(99,'(1X,11E11.3e2)')(Full_left(y,x),x=1,numelem+1)
       enddo
       write(99,*)


C CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C      ! Start SOLUTION loop - solve linear system: Ax=b
C CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

       do i=1,numsteps
   
C        ! Compute new {b} vector
         call calc_b( BplusY,BminusY,gamma,beta,delta_t,
     >            delta_sq_inv,R,D,D_old,numelem,b,
     >            numbcs,imposed_node,imposed_type,
     >            imposed_Fd,Full_left,Right_cur,Right_pre,
     >            traction,length )

C        ! Get old value of D from previous D
         do j=1,numelem+1
           D_old(j) = D(j)
         enddo

C        ! Copy 'Full_left' matrix to A matrix because LU
C        ! decomposition destroys the A matrix each time it
C        ! runs.
         do j=1,numelem+1
           do x=1,numelem+1
             A(x,j) = Full_left(x,j)  
           enddo
         enddo

C        ! Solve the linear system Ax=b
         call ludcmp( A,numelem+1,101,indx,d_info )
         call lubksb( A,numelem+1,101,indx,b )

C        ! Update the new temperature: D = b
         do j=1,numelem+1
           D(j) = b(j)
         enddo

C        ! Output displacements every 'step_incr'th timestep
         if ( mod(i,step_incr) .eq. 0 ) then
           write(77,*) i
           write(77,'(F12.4)') (D(x),x=1,numelem+1)
         endif

       enddo          ! End of SOLUTION loop

       return
       end         ! End of "solution" subroutine



C *******************************************************************
C   
C   calc_b - calculate the right hand side of linear system Ax=b
C 
C *******************************************************************
       subroutine calc_b( BplusY,BminusY,gamma,beta,delta_t,
     >            delta_sq_inv,R,D,D_old,numelem,b,
     >            numbcs,imposed_node,imposed_type,
     >            imposed_Fd,Full_left,Right_cur,Right_pre,
     >            traction,length )

       integer max_nodes,max_elem
       parameter(max_nodes=101,max_elem=100)
       integer i,j,numelem,numbcs,imposed_node(max_nodes), 
     >         imposed_type(max_nodes)

       real*8 BplusY,BminusY,gamma,beta,delta_t,delta_sq_inv,
     >        real_1,R(max_nodes),D(max_nodes),D_old(max_nodes),
     >        b(max_nodes),Full_left(max_nodes,max_nodes),
     >        Right_cur(max_nodes,max_nodes),
     >        Right_pre(max_nodes,max_nodes),
     >        imposed_Fd(max_nodes),traction(max_elem),
     >        length(max_elem)

C      ! Add up right side of Ax=b equation since R(n+1) = R(n)
C      ! So now R(i) will be equivalent to 'b'
       do i=1,numelem+1
         real_1 = 0.0
         do j=1,numelem+1
           real_1 = real_1 + Right_cur(i,j)*D(j) 
     >                     + Right_pre(i,j)*D_old(j)
         enddo
         b(i) = R(i) + real_1
       enddo

C      ! Impose the fixed displacements (i.e. if type = 0)
       do i=1,numbcs
         if ( imposed_type(i) .eq. 0 ) then
           b(imposed_node(i)) = imposed_Fd(i)
         else
           b(imposed_node(i)) = 
     >           b(imposed_node(i)) +
     >           imposed_Fd(i) + 
     >           traction(imposed_node(i-1))*
     >           length(imposed_node(i-1))*0.5
         endif
       enddo


       return
       end         ! End of "calc_b" subroutine 


C *******************************************************************
C   
C   assemble - assemble the global M,K matrises and R vector
C 
C *******************************************************************
       subroutine assemble( K,M,R,E,area,length,density,numelem,
     >       BplusY,BminusY,delta_sq_inv,Full_left,Right_cur,
     >       Right_pre,numbcs,imposed_node,imposed_type,imposed_Fd,
     >       traction,beta )

       integer max_nodes,max_elem
       real*8 one_sixth
       parameter(max_nodes=101,max_elem=100,one_sixth=0.1666667)
       integer i,j,numelem,numbcs,imposed_node(max_nodes),
     >         imposed_type(max_nodes),int_1

       real*8 k_const,m_const,r_const,K(max_nodes,max_nodes),
     >        M(max_nodes,max_nodes),R(max_nodes),E(max_elem),
     >        area(max_elem),length(max_elem),density(max_elem),
     >        BplusY,BminusY,delta_sq_inv,imposed_Fd(max_nodes),
     >        Full_left(max_nodes,max_nodes),traction(max_elem),
     >        Right_cur(max_nodes,max_nodes),
     >        Right_pre(max_nodes,max_nodes),beta
 

C      ! Clear the matrices and vectors initially
       do i=1,numelem+1
         do j=1,numelem+1
           M(i,j) = 0.0
           K(i,j) = 0.0
         enddo
         R(i) = 0.0
       enddo

C      ! Assemble the global matrices and vectors
       do i=1,numelem
C        ! Assemble the [K] matrix
         k_const = E(i)*area(i)/length(i)
         K(i  ,i  ) = K(i  ,i  ) + k_const
         K(i+1,i+1) = K(i+1,i+1) + k_const
         K(i  ,i+1) = K(i  ,i+1) - k_const
         K(i+1,i  ) = K(i+1,i  ) - k_const

C        ! Assemble the [M] matrix
         m_const = density(i)*area(i)*length(i)*one_sixth
         M(i  ,i  ) = M(i  ,i  ) + m_const*2.0
         M(i+1,i+1) = M(i+1,i+1) + m_const*2.0
         M(i  ,i+1) = M(i  ,i+1) + m_const
         M(i+1,i  ) = M(i+1,i  ) + m_const

C        ! Assemble the {R} vector by splitting the applied 
C        ! traction to the two element nodes
         r_const = traction(i)*length(i)*0.5
         R(i  ) = R(i  ) + r_const
         R(i+1) = R(i+1) + r_const
       enddo


C      ! To save calculations, precalculate the constant summed
C      ! matrices: 
       do i=1,numelem+1
         do j=1,numelem+1
           Full_left(i,j) = M(i,j)*delta_sq_inv + beta*K(i,j)
           Right_cur(i,j) = 2.0*M(i,j)*delta_sq_inv-BplusY*K(i,j)
           Right_pre(i,j) = -M(i,j)*delta_sq_inv-BminusY*K(i,j)
         enddo
       enddo

C      ! Loop over imposed boundary displacements
       do i=1,numbcs
         if ( imposed_type(i) .eq. 0 ) then
           int_1 = imposed_node(i)

C          ! Set everything in row to 0.0
           do j=1,numelem+1
             Full_left(int_1,j) = 0.0
           enddo

C          ! Set imposed node to 1.0
           Full_left(int_1,int_1) = 1.0
         endif
       enddo


       return
       end         ! End of "assemble" subroutine


C *******************************************************************
C
C   file_input - read data from input file
C
C *******************************************************************
       subroutine file_input( numsteps,step_incr,delta_t,beta,gamma,
     >            length,density,E,area,numelem,numbcs,imposed_node,
     >            imposed_type,imposed_Fd,traction )

       integer max_nodes,max_elem
       parameter(max_nodes=101,max_elem=100)
       integer numsteps,step_incr,numelem,numbcs,
     >         imposed_node(max_nodes),imposed_type(max_nodes)

       integer i,int_1,int_2,int_3,inlen
       character*20 infile

       real*8 delta_t,beta,gamma,length(max_elem),density(max_elem),
     >        E(max_elem),area(max_elem),imposed_Fd(max_nodes),
     >        total_len,len_small,len_large,traction(max_elem)

       real*8 real_1,real_2,real_3,real_4,real_5,real_6

C      ! Initialize the small and large value of element lengths
       len_small = 99999.9
       len_large = 0.0
       real_6 = 0.0

C      ! Open file for reading
       open(7,file='input_struct',status='old')

 10    format(A100)
 20    format(a20)

C      # File name for data output
       read(7,10)                 ! comments for prog. name
       read(7,20) infile          ! file name

C      ! Prepare for output file
       inlen = 20
       do i = 1,20
         if (infile(i:i).ne.' ') then
           inlen = i
         endif
       enddo
       
C      ! Open problem summary file  
       open(99,file=''//infile(1:inlen)//'.res',status='unknown')
       open(77,file='input_display_struct',status='unknown')

C      # <num tsteps> <tstep incr> <delta_t> <Beta> <Gamma>
       read(7,10)                 ! comments for timestep info
       read(7,*) numsteps,step_incr,delta_t,beta,gamma

C      # <total len(m)> <num elem.>
       read(7,10)                 ! comments for length
       read(7,*) total_len,numelem
         write(77,*) numsteps/step_incr
         write(77,*) numelem+1

C      # Number regions for element sizes
       read(7,10)                 ! comments for # length regions
       read(7,*) int_1

C      # If diff.:<From elem #><To elem #><l><E><rho><A><traction>
       read(7,10)                 ! comments for element lengths
             
       do i=1,int_1
         read(7,*) int_2,int_3,real_1,real_2,real_3,real_4,real_5
             
C        ! Update the smallest element length
         if ( real_1 .lt. len_small ) then
           len_small = real_1
         endif
           
C        ! Update the largest element length
         if ( real_1 .gt. len_large ) then
           len_large = real_1
         endif 

C        ! Assign the size of the elements based on input data
         do j=int_2,int_3
             length(j) = real_1
                  E(j) = real_2
            density(j) = real_3
               area(j) = real_4
	   traction(j) = real_5

C          ! Update overall length calculation   
           real_6 = real_6 + real_1
         enddo     ! end 'do j=int_2,int_3'
       enddo     !  end 'do i=1,int_1'
             
C      ! Check if all the elements add up to the specified length
       if ( abs(real_6 - total_len) .gt. 0.00001 ) then
         write(6,*) "Error(1): Total specified length,",total_len,
     >              " , does not match the sum of the lengths",
     >              real_6  
         stop
       endif

C      # Boundary conditions: <number bcs> (applied at nodes)
       read(7,10)                 ! comments for BCs
       read(7,*) numbcs

C      # BCs: <node #> <type 0=displ / 1=force> <Imposed value>
       read(7,10)                 ! comments for bc values read
       do i=1,numbcs
         read(7,*) imposed_node(i),imposed_type(i),imposed_Fd(i)
       enddo

C      ! Close file
       close(7)

       return
       end           ! End of "file_input" subroutine

























